# -*- coding: UTF-8 -*-

from __future__ import division, absolute_import, print_function

"""
My first attempt to write a Z80 disassembler.
(C)2012,2013 LGB Gábor Lénárt lgb@lgb.hu
Can be used/distributed/modified according to license of GNU/GPL v3 or later.
More information: http://www.gnu.org/licenses/gpl.html
"""


import ctypes
from z80disasmopcode import Z80OpcodeDisassembler



class Z80OpcodeMatrixDump(object):


	def __setmem(self, prefix, optab, opcode):

		for p in range(len(self.mem)): self.mem[p] = 0
		p = 0
		if prefix:
			self.mem[0] = prefix
			p += 1
		if optab:
			self.mem[p] = optab
			p += 1
			if prefix and optab == 0xCB: p += 1 # displacement
		self.mem[p] = opcode


	def __dump(self, res):

		o = []
		if res.cc: o.append(res.cc)
		for a in res.p:
			if a == self.disasm.IXY_IND:	o.append("(" + res.rm + "+dd)")
			elif a == self.disasm.ALSOREG_UNDOC:		o.append(res.cb_pu)
			elif a == self.disasm.DM_B:	o.append("nn")
			elif a == self.disasm.DM_E:	o.append("ee")
			elif a in (self.disasm.DM_W, self.disasm.DM_J):	o.append("nnnn")
			elif a in (self.disasm.DM_L, self.disasm.DM_S):	o.append("(nnnn)")
			elif a == self.disasm.DM_I:	o.append("(nn)")
			elif a == self.disasm.DM_R:	o.append(hex(res.dval))
			elif a == self.disasm.DM_X:	o.append(str(res.dval))
			elif isinstance(a, str): o.append(a)
			else: raise RuntimeError("Unknown type " + str(type(a)) + " value = " + str(a) + " @ op=" + op + " ob=" + hex(res.ob))
		return (res.op, ",".join(o), res.term)


	def dump(self):

		self.mem = bytearray(6)
		self.disasm = Z80OpcodeDisassembler(
			codehint = None,
			first_byte = 0,
			last_byte = len(self.mem) -1,
			mem = self.mem
		)
		o = []
		for byte in range(256):
			o1 = []
			for row in ((0, 0), (0xDD, 0), (0, 0xCB), (0xDD, 0xCB), (0, 0xED)):
			#for row in ((0, 0), (0xDD, 0), (0, 0xCB), (0xDD, 0xCB)):
				if byte in (0xDD, 0xFD, 0xCB, 0xED) and row[1] == 0:
					oo = ("PREFIX", hex(byte)[-2:].upper(), False)
				else:
					self.__setmem(row[0], row[1], byte)
					try:
						res = self.disasm.analyze_opcode(0)
					except self.disasm.unknownZ80OpCodeError:
						oo = ("UNDEF", "", True)
					else:
						oo = self.__dump(res)
					#if len(o1) == 1 and oo == o1[0]: oo = ("", "")
				o1.append(oo)
			o.append(o1)
		return o



class Z80Disassembler(object):

	DISASM_COMMENT = "; Disassembled by LGB's two pass Z80/Enterprise disassembler"
	HEXDUMP_WIDTH = 0x10
	ASCII_STR_MINSIZE =  5
	ASCII_STR_MAXSIZE = 64
	EP_IOREGS = {
		0x00: "Serial card 'Meszaros'",
		0x01: "Serial card 'Meszaros'",
		0x10: "EXDOS WD177x write command register / read status register",
		0x11: "EXDOS WD177x track register read/write",
		0x12: "EXDOS WD177x sector register read/write",
		0x13: "EXDOS WD177x data register read/write",
		0x14: "EXDOS WD177x [echo of port 10h] write command register / read status register",
		0x15: "EXDOS WD177x [echo of port 11h] track register read/write",
		0x16: "EXDOS WD177x [echo of port 12h] sector register read/write",
		0x17: "EXDOS WD177x [echo of port 13h] data register read/write",
		0x18: "EXDOS card status/control register",
		0x19: "EXDOS card [echo of port 18h] status/control register",
		0x1A: "EXDOS card [echo of port 18h] status/control register",
		0x1B: "EXDOS card [echo of port 18h] status/control register",
		0x1C: "EXDOS card [echo of port 18h] status/control register",
		0x1D: "EXDOS card [echo of port 18h] status/control register",
		0x1E: "EXDOS card [echo of port 18h] status/control register",
		0x1F: "EXDOS card [echo of port 18h] status/control register",
		0x20: "Turbo EXDOS speed switching",
		0x40: "Spectrum emulator",
		0x41: "Spectrum emulator",
		0x42: "Spectrum emulator",
		0x43: "Spectrum emulator",
		0x44: "Spectrum emulator, write: enable emulator (bit 7)",
		0x7E: "CMOS RTC/memory register select",
		0x7F: "CMOS RTC/memory register read/write",
		0x80: "Nick FIXBIAS",
		0x81: "Nick border color",
		0x82: "Nick LPL",
		0x83: "Nick LPH",
		0x84: "Nick [echo of port 80h] FIXBIAS",
		0x85: "Nick [echo of port 81h] border color",
		0x86: "Nick [echo of port 82h] LPL",
		0x87: "Nick [echo of port 83h] LPH",
		0x88: "Nick [echo of port 80h] FIXBIAS",
		0x89: "Nick [echo of port 81h] border color",
		0x8A: "Nick [echo of port 82h] LPL",
		0x8B: "Nick [echo of port 83h] LPH",
		0x8C: "Nick [echo of port 80h] FIXBIAS",
		0x8D: "Nick [echo of port 81h] border color",
		0x8E: "Nick [echo of port 82h] LPL",
		0x8F: "Nick [echo of port 83h] LPH",
		0xA0: "Dave tone channel 0 frequency low byte",
		0xA1: "Dave tone channel 0 frequency high nybble + params",
		0xA2: "Dave tone channel 1 frequency low byte",
		0xA3: "Dave tone channel 1 frequency high nybble + params",
		0xA4: "Dave tone channel 2 frequency low byte",
		0xA5: "Dave tone channel 2 frequency high nybble + params",
		0xA6: "Dave noise frequency, polynominal counter + misc",
		0xA7: "Dave sync, D/A mode, and interrupt selection",
		0xA8: "Dave tone channel 0 left volume",
		0xA9: "Dave tone channel 1 left volume",
		0xAA: "Dave tone channel 2 left volume",
		0xAB: "Dave noise channel left volume",
		0xAC: "Dave tone channel 0 right volume",
		0xAD: "Dave tone channel 1 right volume",
		0xAE: "Dave tone channel 2 right volume",
		0xAF: "Dave noise channel right volume",
		0xB0: "Dave MMU reg, page 0 (0000-3FFF) selector",
		0xB1: "Dave MMU reg, page 1 (4000-7FFF) selector",
		0xB2: "Dave MMU reg, page 2 (8000-BFFF) selector",
		0xB3: "Dave MMU reg, page 3 (C000-FFFF) selector",
		0xB4: "Dave enable/reset interrupt sources/latches",
		0xB5: "Dave active low strobe on WR0/RD0 (keyboard row select/read)",
		0xB6: "Dave active low strobe on WR1/RD1 ",
		0xB7: "Dave active low strobe on WR2/RD2",
		0xBF: "Dave sysconfig register (wait states, clkdiv)",
		0xEC: "IDE data register low byte",
		0xED: "IDE data register high byte",
		0xEE: "IDE command register port 1",
		0xEF: "IDE command register port 0",
		0xF0: "External DAC write",
		0xF1: "External DAC write",
		0xF2: "External DAC write",
		0xF3: "External DAC write",
		0xFE: "Spectrum emulator",
		0xFF: "Spectrum emulator",
	}
	EXOS_RST = 0x30
	# Syntax of EXOS_FUNCS dict:
	# key: exos call code
	# value is tupple:
	# first element: text description about the function
	# second element: a dict, describing input parameters (register or register pair!)
	#	as keys, values are descriptions, if value starts with '@' then memory pointer
	# third elemenet: list of registers (not pairs!) whose original values will be PRESERVED, default: A, DE, BC, HL is destroyed
	EXOS_FUNCS = {
		 #0: "System reset",
		 1: ("Open channel", {"DE": "@device/filename string", "A": "channel number"}, []),
		 2: ("Create channel", {"DE": "@device/filename string", "A": "channel number"}, []),
		 3: ("Close channel", {"A": "channel number"}, ["A"]),
		 4: ("Destroy channel", {"A": "channel number"}, ["A"]),
		 5: ("Read character", {}, []),
		 6: ("Read block", {}, []),
		 7: ("Write character", {}, []),
		 8: ("Write block", {}, []),
		 9: ("Channel read status", {}, []),
		10: ("Set and reset channel status", {}, []),
		11: ("Special function", {}, []),
		16: ("Read/write/toggle EXOS variable", {}, []),
		17: ("Capture channel", {}, []),
		18: ("Redirect channel", {}, []),
		19: ("Set default device name", {}, []),
		20: ("Return system status", {}, []),
		21: ("Link device", {}, []),
		22: ("Read EXOS boundary", {}, []),
		23: ("Set user boundary", {}, []),
		24: ("Allocate segment", {}, []),
		25: ("Free segment", {}, []),
		26: ("Scan system extensions", {}, []),
		27: ("Allocate channel buffer", {}, []),
		28: ("Explain error code", {}, []),
		29: ("Load module", {}, []),
		30: ("Load relocatable module", {}, []),
		31: ("Set time", {}, []),
		32: ("Read time", {}, []),
		33: ("Set date", {}, []),
		34: ("Read date", {}, [])
	}


	def __hex_comment(self, bs, pc):

		return self.hexword(pc) + " " + " ".join([self.hexbyte(a) for a in bs]) + " \"" + "".join([(chr(a) if self.ascii_pc[a] else "?") for a in bs]) + "\""

	def __flush_string(self, string, pc):

		if len(string) == 0:
			return []
		pc -= len(string)
		if len(string) >= self.ASCII_STR_MINSIZE:
			s2s = (str(string.decode()) if isinstance(string, bytearray) else string)
			return ["\tDB \"" + self.stringize(s2s.replace('\\', '\\\\').replace('"', '\\"'), "string") + "\"\t" +
				self.stringize("; " + self.__hex_comment(string, pc), "comment")
			]
		else:
			return ["\tDB " + ",".join([self.stringize("$" + self.hexbyte(a), "hex") for a in string]) + "\t" + self.stringize("; " + self.__hex_comment(string, pc), "comment")]


	def __disasm(self, start, label_hinting = None):

		codehint = [None] * len(self.mem)
		dasm = Z80OpcodeDisassembler(
			first_byte = self.first_byte,
			last_byte = self.last_byte,
			codehint = codehint,
			mem = self.mem
		)
		# PASS-1 of disassembly
		queue, iterations, clabels, dlabels, done = start[:], 0, {start[0]: "PROGRAM_ENTRY_POINT"}, {}, set()
		queue2, pass1a, anacomment, refs = [], False, {}, {}
		for a in start[1:]: clabels[a] = "CODE_HINTING_ENTRY_" + self.hexword(a)
		while len(queue): # repeat till there is at least one address in the queue
			iterations += 1
			dasm.reg.reset() # start with uninitialized registers
			pc, queue = queue[0], queue[1:]
			anastart, rp_mod_at = iterations, {}
			start_pc = pc
			while True: # endless loop, code will "break" it if there is end of linear code flow, or other problems detected
				done.add(pc)
				if pass1a and pc in dlabels:
					anacomment[pc] = "stop analyzing: in fall-back mode data label found, revoking further CODE mode"
					#for a in range(start_pc, pc + 1): codehint[a] = None
					#anacomment[start_pc] = "stop analyzing: at PC=" + hex(pc) + " problem is detected in fall-back CODE mode, revoking CODE from here till that byte"
					break # this seems to be data, not code, dude!
				try:
					res = dasm.analyze_opcode(pc) # this is the Z80 raw-level opcode disassembler call
				except dasm.alreadyAnalyzedZ80OpCodeError as a:
					anacomment[pc] = "stop analyzing: already analyzed code at iteration #" + str(iterations) + ": " + str(a)
					break # break the loop: already analyzed code
				except dasm.unknownZ80OpCodeError as a:
					anacomment[pc] = "stop analyzing: unknown Z80 opcode at iteration #" + str(iterations) + ": " + str(a)
					#for a in range(start_pc, pc + 1): codehint[a] = None
					#anacomment[start_pc] = "stop analyzing: at PC=" + hex(pc) + " problem is detected in CODE mode, revoking CODE from here till that byte"
					break # break the loop if unknown opcode
				except dasm.outOfBoundZ80OpCodeError as a:
					anacomment[pc] = "stop analyzing: out-of-bound opcode #" + str(iterations) + ": " + str(a)
					break # break the loop if out-of-bound code

				codehint[pc] = res # store in our array
				if pass1a and anastart: anacomment[pc] = "fall-back code analyzation starting point [maybe incorrect, and this is data!]"
				res.anastart, anastart = anastart, 0
				m = [a for a in res.p if isinstance(a, int)]  # search for special opcode parameters
				if m:
					m = m[0]
					if m == dasm.DM_W:
						# register pair change (with constant) will be booked [used by EXOS call analyzer part]
						if len(res.p[0]) == 2: rp_mod_at[res.p[0]] = pc
						res.r16label = False
					elif m in (dasm.DM_J, dasm.DM_E):
						if res.dval not in done and res.dval not in queue:
							# in case of jump instruction, add the jump address to the queue if not already there, or done already
							# in a way that queue is sorted to try to maximize the possibility to found more linear code flows
							queue = sorted(queue + [res.dval])
						if res.dval not in clabels:
							# Create meaningful labels for jumps
							if m == dasm.DM_E: a = "NEAR"  # one byte signed relative jump will have NEAR_* labels
							elif res.op == "CALL": a = "SUB" # CALL ops will have SUB_* labels
							else: a = "FAR" # two byte non-CALL (JP) ops will have FAR_* labels
							clabels[res.dval] = a + "_" + self.hexword(res.dval)
						if res.dval in refs: refs[res.dval].append(pc)
						else: refs[res.dval] = [pc]
					elif m in (dasm.DM_L, dasm.DM_S):
						# direct memory read/write (with constant addresses) are dereferenced with lables as well
						if res.dval not in dlabels:
							dlabels[res.dval] = "DATA_" + self.hexword(res.dval)
						if res.dval in refs: refs[res.dval].append(pc)
						else: refs[res.dval] = [pc]
					elif m == dasm.DM_R: # RST check for EXOS calls
						if res.dval == self.EXOS_RST and pc < self.last_byte: # RST 0x30 is the EXOS call, PC check is for an abnormal situation where last byte is RTS 0x30 in the program without the exos func code byte then
							exos = self.mem[pc + 1] # read the next byte: exos function code
							res.op, res.p, res.dval, res.pc = "EXOS", [dasm.DM_B], exos, pc + 2 # "fake" an EXOS opcode!
							if exos in self.EXOS_FUNCS:
								res.comment.append("EXOS call (" + hex(exos) + "): " + self.EXOS_FUNCS[exos][0])
							else:
								res.comment.append("EXOS call, unknown function: " + hex(exos))
							if dasm.reg.zregs["D"] is not None and dasm.reg.zregs["E"] is not None:
								de = (dasm.reg.zregs["D"] << 8) | dasm.reg.zregs["E"]
								res.comment.append("DE is known! :-) -> " + hex(de))
								#dlabels[de] = "EXOS_STR_" + self.hexword(de)
								if "DE" in rp_mod_at:
									dlabels[de] = "EXOS_STR_" + self.hexword(de)
									codehint[rp_mod_at["DE"]].r16label = True
									codehint[rp_mod_at["DE"]].comment.append("EXOS filename string for EXOS call " + hex(exos) + " at " + hex(pc))
									if de in refs: refs[de].append(pc)
									else: refs[de] = [pc]
							if dasm.reg.zregs["A"] is not None:
								res.comment.append("A=" + hex(dasm.reg.zregs["A"]))
							# Reset waiting for register pairs [TODO: maybe not always should be done?!]
							rp_mod_at = {}
							# Registers lost
							for a in ("D", "E", "A"): dasm.reg.zregs[a] = None
				if res.term:
					res.comment.append("Opcode breaks the linear code flow")
					if res.pc not in queue2 and res.pc not in done:
						queue2 = sorted(queue2 + [res.pc])
					break # Ok, end of linear code flow, TODO: maybe we should consider to disassemble further??
				pc = res.pc # continue with the next opcode
			if len(queue) == 0: # fall-back mode: after breakge of linear code analyzation, we try to continue at the next address anyway!
				queue, queue2, pass1a = queue2, [], True
		del queue, queue2, done, res
		# PASS-2/A eliminate labels which cannot be reached (middle of an opcode, out-of-bound, etc)
		# "reachable" is a list which will contain all of the reachable label positions
		a, reachable = self.first_byte, []
		while a <= self.last_byte:
			reachable.append(a)
			if codehint[a]: a = codehint[a].pc
			else: a += 1
		for a in list(clabels): # eliminate impossible code labels
			if a < self.first_byte or a > self.last_byte: del clabels[a]
			elif codehint[a] is None: del clabels[a]
			elif a not in reachable: del clabels[a]
		#print(dlabels)
		#print(dlabels.keys())
		#print(dlabels.keys()[:])
		for a in list(dlabels): # eliminate impossible data labels
			if a not in reachable: del dlabels[a]
		self.dasminfo.append(str(len(reachable)) + " locations in the image")
		self.dasminfo.append(str(len(dlabels)) + "/" + str(len(clabels)) + " data/code labels created")
		self.dasminfo.append(str(iterations) + " iterations used to disassemble")
		del reachable
		# PASS-2/B actual disassembly follows
		o, pc, string, om, xrefs = [], self.first_byte, bytearray(), None, []
		for a in refs.values(): xrefs += a
		self.dasminfo.append(str(len(xrefs)) + " xref points found")
		while pc <= self.last_byte:
			res, c = codehint[pc], []
			if string and (len(string) > self.ASCII_STR_MAXSIZE or res or pc in clabels or pc in dlabels or pc in anacomment or pc in refs):
				o += self.__flush_string(string, pc)
				string = bytearray()
			x = (self.anchor(self.hexword(pc)) if pc in xrefs else "")
			if pc in clabels:
				if pc in dlabels: o.append("\t" + self.stringize("; Warning: both of data and code labels, maybe this is not code, but data?", "comment"))
				o.append(self.anchor(clabels[pc]) + self.stringize(clabels[pc], "name") + ":\t\t" + self.stringize("; code label", "comment")) # emit a code label here, if there is one
			if pc in dlabels:
				o.append(self.anchor(dlabels[pc]) + self.stringize(dlabels[pc], "name") + ":\t\t" + self.stringize("; data label", "comment")) # emit a data label here, if there is one
			if pc in anacomment: o.append("\t" + self.stringize("; " + anacomment[pc], "comment"))
			if pc in refs:
				o.append("\t" + self.stringize("; Referenced at ", "comment") + " ".join([self.stringize_link(self.hexword(a), self.hexword(a)) for a in refs[pc]]))
			if res:
				if om != "code":
					o.append("\t" + self.stringize("; Disassembler: switching to CODE mode", "comment"))
					om = "code"
				#c.append(self.hexword(pc) + " " + self.hexdump(self.mem[pc : res.pc]))
				c.append(self.__hex_comment(self.mem[pc : res.pc], pc)) # was an error, res.pc at the second arg!
				c += res.comment
				if res.anastart: c.append("Code analyzation starting point at iteration #" + str(res.anastart) + " of " + str(iterations))
				p = ([self.stringize(res.cc, "token")] if res.cc else [])
				for a in res.p:
					if a in (dasm.DM_J, dasm.DM_E):
						p.append(self.stringize_link(clabels[res.dval], clabels[res.dval], "name") if res.dval in clabels else self.stringize("$" + self.hexword(res.dval), "hex"))
					elif a in (dasm.DM_L, dasm.DM_S):
						p.append("(" + (self.stringize_link(dlabels[res.dval], dlabels[res.dval], "name") if res.dval in dlabels else self.stringize("$" + self.hexword(res.dval), "hex")) + ")")
					elif a == dasm.DM_R:
						p.append(self.stringize("$" + self.hexbyte(res.dval), "hex"))
					elif a == dasm.DM_I:
						if res.dval in self.EP_IOREGS:
							c.append("I/O on port " + hex(res.dval) + ": " + self.EP_IOREGS[res.dval])
						p.append("(" + self.stringize("$" + self.hexbyte(res.dval), "hex") + ")")
					elif a == dasm.DM_X:
						p.append(str(res.dval))
					elif a == dasm.DM_W:
						if res.r16label and res.dval in dlabels:
							p.append(self.stringize_link(dlabels[res.dval], dlabels[res.dval], "name"))
						else:
							p.append(self.stringize("$" + self.hexword(res.dval), "hex"))
					elif a == dasm.DM_B:
						p.append(self.stringize("$" + self.hexbyte(res.dval), "hex"))
					elif isinstance(a, int):
						raise RuntimeError("__disasm() failed, unknown DM=" + str(a) + " at PC=" + hex(pc))
					elif a == dasm.ALSOREG_UNDOC:
						p.append(res.cb_pu)
					elif a == dasm.IXY_IND:
						p.append("(" + res.rm + ("-" if res.disp < 0 else "+") + self.stringize("$" + self.hexbyte(abs(res.disp)), "hex") + ")")
					else:
						p.append(a)
				o1 = "\t" + self.stringize(res.op, "token") + (4 - len(res.op)) * " " + " " + ",".join(p)
				# Ok, the next opcode follows
				pc = res.pc
			else: # data mode
				if om != "data":
					o.append("\t" + self.stringize("; Disassembler: switching to DATA mode", "comment"))
					om = "data"
				a = self.mem[pc]
				if self.ascii_pc[a]: # ascii_pc is a bool table with char pos where selected EP chset and US-ASCII on PC (low 7 bit of unicode) match
					string.append(a)
					o1 = None # no output for now ...
				else:
					if string:
						o += self.__flush_string(string, pc)
						string = bytearray()
					o1 = "\tDB " + self.stringize("$" + self.hexbyte(a), "hex") + "\t" + self.stringize("; " + self.__hex_comment([a], pc), "comment")

					#if self.ascii_pc[a]:
					#	o1 += " '" + chr(a) + "'"
				pc += 1 # next byte
			if o1:
				if c:
					o1 += "\t" + self.stringize("; " + c[0], "comment")  # first comment (normally PC and bytes) is after the opcode
					c = c[1:]
				for a in c: o.append("\t" + self.stringize("; " + a, "comment")) # additional comments before the opcode line
				o.append(x + o1) # add the result to the disassembly list

		return "\n".join(o + self.__flush_string(string, pc) + ["END_OF_PROGRAM:\t\t; Do not put anything after this!"]) # flush is used to really flush "string" if there is any left at the end of the prg

	def hexword(self, a):

		return ("000" + hex(a)[2:])[-4:].upper()

	def analyze_ml_prg(self, prg, size, load, start, html = False, debug = False, dasm = False, preprint = None):

		if load < 0: raise RuntimeError("Negative load address is insane.")
		for a in start:
			if a < load: raise RuntimeError("Start/code hinting address(es) cannot be smaller than load address.")
		self.dasminfo = []
		#self.mem = bytearray([0]*0x10000) # Simulate 64K RAM. Note: disassembler (at least my one ...) can't be so smart to detect memory mapping etc :(
		if load > 0xFFFF: raise RuntimeError("Load address exceeds 64K")
		self.last_byte = load + size - 1
		if size <= 0: raise RuntimeError("Null (or negative?) sized program?")
		if self.last_byte > 0xFFFF: raise RuntimeError("Program does not fit into 64K (load+size)")
		for a in start:
			if a > self.last_byte: raise RuntimeError("Start/code hinting address(es) cannot be after the last byte of the loaded program")
		self.first_byte = load
		#prg = prg[:size]
		if len(prg) != size: raise RuntimeError("Program is shorter than the specified size")
		#self.mem[load:load + size] = prg
		self.mem = bytearray(load) + bytearray(prg[:size])
		self.dasminfo.append("Program load address: " + hex(load) + ", size is " + hex(size) + ", last byte is " + hex(self.last_byte))
		self.dasminfo.append("Program start address (+ code hint points): " + " ".join(map(hex, start)))
		self.dasminfo.append("ML analyzation mode: " + ("disassembly" if dasm else "hex dump (you can try to disassemble it with the -dasm switch)"))
		o = (preprint + "\n" if preprint else "")
		if dasm:
			# Disassembly was requested!
			o = self.DISASM_COMMENT + "\n" + o + "\n\tMACRO EXOS n ; by LGB\n\tRST " + hex(self.EXOS_RST) + "\n\tDB n\n\tENDMACRO\n\n" + self.__disasm(start)
		else:
			# No disassembly was requested, continue with normal hex dump
			a = load
			while a <= self.last_byte:
				s = self.mem[a:a + self.HEXDUMP_WIDTH]
				o += self.stringize(self.hexword(a), "token") + "  " + self.stringize(self.hexdump(s), "hex") + " " * (self.HEXDUMP_WIDTH - len(s)) * 3 + "  " + self.stringize(s) + "\n"
				a += self.HEXDUMP_WIDTH
		return o



if __name__ == "__main__": raise RuntimeError("Do not run this file directly, it's not the main 'executable'!")
