# -*- coding: UTF-8 -*-

from __future__ import division, absolute_import, print_function

"""
My first attempt to write a Z80 disassembler, please be patient :)
(C)2012,2013 LGB Gábor Lénárt lgb@lgb.hu
Can be used/distributed/modified according to license of GNU/GPL v3 or later.
More information and the license text itself: http://www.gnu.org/licenses/gpl.html

Notes about the disassembler:
YES, it would be easier to generate fixed tables, rather than "decoding"
each opcodes on-the-fly, but this is much more fun, and it has got the
advantage to directly fixing bugs, instead of regenerating tables first, etc.
Also, information got from ML code directly helps some analysis, like limited
"following" mode for register values, etc.

This disassembler itself does not generate too much, it's only the
skeleton (though the major part of the problem to solve), another
python source will use and actually convert the information into text
created by class Z80OpcodeDisassembler.
"""

import ctypes


class Z80Registers(object):

	zregnames16 = {
		"AF": ("A", "F"), "BC": ("B", "C"), "DE": ("D", "E"), "HL": ("H", "L"),
		"IX": ("IXH", "IXL"), "IY": ("IYH", "IYL"), "SP": ("SPH", "SPL"),
		"PC": ("PCH", "PCL"), "AF'": ("A'", "F'"), "BC'": ("B'", "C'"),
		"DE'": ("D'", "E'"), "HL'": ("H'", "L'")
	}
	F, F2 = [None] * 8, [None] * 8

	def __init__(self): self.reset()

	def reset(self):
		self.zregs = {}
		for a in self.zregnames16.values():
			self.zregs[a[0]], self.zregs[a[1]] = None, None

	def set_f_bit(self, bit, val): pass

	def get_f_bit(self, bit, val): pass

	def ex8(self, r1, r2):

		if r1 in self.zregs and r2 in self.zregs:
			self.zregs[r1], self.zregs[r2] = self.zregs[r2], self.zregs[r1]

	def ex16(self, r1, r2):

		if r1 in self.zregnames16 and r2 in self.zregnames16:
			self.ex8(self.zregnames16[r1][0], self.zregnames16[r2][0])
			self.ex8(self.zregnames16[r1][1], self.zregnames16[r2][1])

	def set8(self, r, n):

		if r in self.zregs: self.zregs[r] = (None if n is None else n & 255)

	def set8_more(self, rs, n):

		for a in rs: self.set8(n)

	def set16(self, r, n):

		if r in self.zregnames16:
			self.set8(self.zregnames16[r][0], (None if n is None else n >> 8))
			self.set8(self.zregnames16[r][1], n)

	def copy8(self, target, source):

		if target in self.zregs:
			self.zregs[target] = (self.zregs[source] if source in self.zregs else None)

	def copy16(self, target, source):

		if target in self.zregnames16:
			self.zregs[self.zregnames16[target][0]] = self.zregs[self.zregnames16[source][0]] if source in self.zregnames16 else None
			self.zregs[self.zregnames16[target][1]] = self.zregs[self.zregnames16[source][1]] if source in self.zregnames16 else None

	def add8(self, r, n): #: should be used only for INC/DEC so n=(1, -1) ...

		if r in self.zregs and self.zregs[r] is not None:
			self.zregs[r] = ctypes.c_uint8(self.zregs[r] + n).value



class Z80OpcodeDisassembler(object):

	cc_tab = ("NZ", "Z", "NC", "C", "PO", "PE", "P", "M")
	r8_tab = ("B", "C", "D", "E", "H", "L", "(HL)", "A")
	alu_tab = ("ADD", "ADC", "SUB", "SBC", "AND", "XOR", "OR", "CP")
	r16_tab  = ("BC", "DE", "HL", "SP")
	r16_tab2  = ("BC", "DE", "HL", "AF")

	IXY_IND = "IXYIND" #: parameter is (IX+d) or (IY+d): reg name is in res.rm, d is in res.disp
	ALSOREG_UNDOC = "ALSOREGUNDOC" #: undocumented "also into reg" feature in CB ops, reg name in res.cb_pu
	# the rest of these are more "standard": res.dval is the value itself, and the exact constant shows how to interpret the value
	# to learn more, read method OP()
	DM_B = 1	#: byte value as data
	DM_W = 2	#: word value as data
	DM_E = 3	#: PC relative jump address (converted to "real" one), for jr, djnz [signed byte]
	DM_L = 4	#: word value as address to load data from, usually this must be replaced with brackets, ie: (0x34FE)
	DM_S = 5	#: word value as address to store data to, about the same as DM_L, just for storing, not loading
	DM_J = 6	#: jump address for jp, call .... [word]
	DM_I = 7	#: I/O address [byte, IN/OUT ops]
	DM_R = 8	#: RST number
	DM_X = 9	#: bit index for CB opcodes

	class outOfBoundZ80OpCodeError(BaseException): pass
	class alreadyAnalyzedZ80OpCodeError(BaseException): pass
	class unknownZ80OpCodeError(BaseException): pass

	def __init__(self, mem, first_byte, last_byte, codehint):

		# Some sanity checks
		if first_byte > last_byte:
			raise RuntimeError("first_byte is higher than last_byte")
		if first_byte < 0:
			raise RuntimeError("first_byte is negative")
		if last_byte > 0xFFFF:
			raise RuntimeError("last_byte exceeds 64K")
		if last_byte >= len(mem):
			raise RuntimeError("last_byte exceeds the given 'mem' bytearray's size")
		if codehint is not None and len(codehint) != len(mem):
			raise RuntimeError("size of 'codehint' and 'mem' does not matches")
		# Store parameters
		self.first_byte = first_byte
		self.last_byte = last_byte
		self.codehint = codehint
		self.mem = mem
		# Registers
		self.reg = Z80Registers()


	def opcb(self, signed = False):

		"""This method reads one byte of opcode or its part,
		raising exceptions in some problematic cases (which should be handled by the frontend!)."""

		if self.res.pc < self.first_byte or self.res.pc > self.last_byte:
			raise self.outOfBoundZ80OpCodeError("Out-of-bound code, end of image before decoding the op")
		if self.codehint is not None and self.codehint[self.res.pc]:
			raise self.alreadyAnalyzedZ80OpCodeError("Opcode (or part of it) has been already analyzed")
		b = self.mem[self.res.pc]
		self.res.pc += 1
		return (ctypes.c_byte(b).value if signed else b)

	def opcw(self, signed = False):

		w = self.opcb() # Note: even if we need signed, we fetch bytes as unsigned and covert the word at the end!
		w |= self.opcb() << 8 # Note2: we uses byte fetch so no need to check here for limits/etc. But then it's important to fetch bytes in the good order!!
		return (ctypes.c_int16(w).value if signed else w)

	def opcdisp(self):

		self.res.disp = self.opcb(True) # fetch a _SIGNED_ byte as displacement


	def OP(self, op, par = None, cc = None, term = None):

		if self.res.optab not in (0, 0xED):
			raise RuntimeError("You can't call OP() with this optab: " + self.errbase)
		self.res.op, self.res.cc = op, cc # store op name and possible condition
		if term == True:
			if cc is None and op != "DJNZ": self.res.term = True
			else: self.res.mayterm = True
		if par and self.res.prefix and self.res.optab == 0 and self.res.ob != 0xEB:
			try: a = par.index("(HL)")
			except ValueError: a = -1
			if a >= 0:
				self.opcdisp() # fetch displacement
				par[a] = self.IXY_IND # signal the IX/IY indexed with displacement - (IX+d) or (IY+d) - situation for the disassembler frontend
			else:
				for a in range(len(par)): # convert: HL -> IX/IY, H -> IXH/IYL, L -> IYH/IYL
					if par[a] == "HL": par[a] = self.res.rm
					elif par[a] in ("L", "H"): par[a] = self.res.rm + par[a]
		if par:
			if not isinstance(par, list):
				raise RuntimeError("Not list for par in OP() but: " + str(type(par)) + " at " + self.errbase)
			#par = list(par)
			# check value needed to be fetched
			m = None
			for a in range(len(par)):
				if isinstance(par[a], int):
					if m: raise RuntimeError("A DM code (" + str(m) + ") is already given in OP(): " + self.errbase)
					m = par[a]
				elif not isinstance(par[a], str):
					raise RuntimeError("OP() param #" + str(a) + " is unknown type: " + str(type(par[a])) + " at " + self.errbase)
			if m in (self.DM_B, self.DM_I):
				self.res.dval = self.opcb()
			elif m in (self.DM_W, self.DM_L, self.DM_S, self.DM_J):
				self.res.dval = self.opcw()
			elif m == self.DM_E:
				# we do this in two steps to avoid a problematic point: opcb() increments res.pc, so it's very important to do opcb() call and res.pc addition in the right order!
				self.res.dval = self.opcb(True)
				self.res.dval += self.res.pc
				if self.res.dval < 0 or self.res.dval > 0xFFFF:
					raise self.unknownZ80OpCodeError("Valid op, but PC-relative JR/DJNZ address over-/underflow")
			elif m not in (None, self.DM_R, self.DM_X): # important note about DM_X, DM_R: you must set value by yourself!
				raise RuntimeError("Unknown DataMode " + str(m) + " reference for opcb=" + hex(self.res.ob) + " for optab " + hex(self.res.optab))
			self.res.p = par


	class __DisAsmData(object):

		def __init__(self, pc):

			self.pc = self.pc_old = pc
			self.p, self.comment, self.prefix = [], [], 0
			self.term = self.mayterm = False
			self.disp = self.cc = self.dval = self.op = self.optab = self.cb_pu = None


	def analyze_opcode(self, pc):

		res = self.res = self.__DisAsmData(pc)

		# Fetch opcode bytes: if DD, FD set prefix but continue to fetch. Ready if non-DD/FD opcode byte is fetched
		# Note: it's an undocumented feature of Z80: you can specify large sequence of DD/FD bytes, then the last counts,
		# and Z80 will fetch bytes till a non-DD/FD opcode byte is found (and interrupts are disabled till that!)
		# By contrast, ED/CB is taken immediately (that's why they're not handled inside this loop)
		while res.optab is None:	# repeat till optab is set to a value in the loop
			o = self.opcb()
			# Note: FD/DD prefixes (not CB/ED!) can be repeated!
			if o in (0xDD, 0xFD):	# 0xDD/0xFD prefix, memorize prefix, but continue reading bytes (no optab is set!)
				if res.prefix and not res.comment: res.comment = ["Unusual sequence of DD/FD prefixes"]
				res.prefix, res.rm = o, "IX" if o == 0xDD else "IY"
			elif o == 0xCB:	# 0xCB opcode, fetch displacement (disp) if FD/DD prefixed, fetch the opcode into "o", set optab to 0xCB
				if res.prefix: self.opcdisp() # we need displacement if DD/FD prefixed
				res.optab, o = o, self.opcb() # set optab to CB, and fetch the opcode
			elif o == 0xED:	# 0xED opcode, fetch opcode into "o" (ED ignores previous DD/FD prefix!), set optab to 0xED
				if res.prefix: res.comment.append("ED opcodes ignores FD/DD prefixes")
				res.optab, o = o, self.opcb() # set optab to ED, and fetch the opcode
			else:		# other byte, set optab to 0, "o" contains the already fetched value as the opcode
				res.optab = 0

		res.ob = o # opcode byte is also represented in "res"
		self.errbase = "opcode " + hex(o) + " [optab=" + hex(res.optab) + "] at PC=" + hex(res.pc - 1) # used by error handlers, exception raisers
		# These variables help to decode opcodes
		x, y, z = o >> 6, (o >> 3) & 7, o & 7
		p, q = y >> 1, y & 1

		if res.optab == 0xCB: # CB opcode series (w/wo DD/FD prefixes!), still as you can see: CB opcodes are nice and quite easy to decode :)

			if x == 0:
				res.op = ("RLC", "RRC", "RL", "RR", "SLA", "SRA", "SLL", "SRL")[y]
			else:
				res.op, res.p, res.dval = ("BIT", "RES", "SET")[x - 1], [self.DM_X], y
			if res.prefix:
				res.p += [self.IXY_IND]
				if z != 6:
					res.cb_pu = self.r8_tab[z]
					res.p += [self.ALSOREG_UNDOC]
			else:
				res.p += [self.r8_tab[z]]

		elif res.optab == 0xED: # ED opcodes are nice, since DD/FD prefixes are simply ignored. Yay!

			if x == 1:
				if z == 0:
					if y == 6: res.op, res.p = "IN", ["(C)"]
					else: res.op, res.p = "IN", [self.r8_tab[y], "(C)"]
				elif z == 1:
					if y == 6: res.op, res.p = "OUT", ["(C)", "0"]
					else: res.op, res.p = "OUT", ["(C)", self.r8_tab[y]]
				elif z == 2:
					res.op, res.p = "ADC" if q else "SBC", ["HL", self.r16_tab[p]]
				elif z == 3:
					if q:
						self.OP("LD", [self.r16_tab[p], self.DM_L])
					else:
						self.OP("LD", [self.DM_S, self.r16_tab[p]])
				elif z == 4:
					res.op = "NEG"
				elif z == 5:
					res.op = "RETI" if y == 1 else "RETN"
				elif z == 6:
					res.op, res.p = "IM", [("0", "0/1", "1", "2")[y & 3]]
				elif z == 7:
					if y < 4: res.op, res.p = "LD", (["I", "A"], ["R", "A"], ["A", "I"], ["A", "R"])[y]
					else: res.op = ("RRD", "RLD", "NOP", "NOP")[y - 4]
			elif x == 2 and z < 4 and y > 3:
				res.op = ( # ["LD", "CP", "IN", "OT"][z] + ["I", "D", "IR", "DR"][y - 4]
					("LDI",	"CPI",	"INI",	"OUTI"),
					("LDD",	"CPD",	"IND",	"OUTD"),
					("LDIR","CPIR",	"INIR",	"OTIR"),
					("LDDR","CPDR",	"INDR",	"OTDR")
				)[y - 4][z]
			else:
				raise self.unknownZ80OpCodeError("Unknown ED opcode")


		elif res.optab == 0: # "normal" opcode table (no ED/CB) but possibly DD/FD prefixed

			if x == 0:
				if z == 0:
					if y == 0:
						self.OP("NOP")
					elif y == 1:
						self.OP("EX", ["AF", "AF'"])
						self.reg.ex16("AF", "AF'")
					elif y == 2:
						self.OP("DJNZ", [self.DM_E], term = True)
						self.reg.set8("B", None)
					elif y == 3:
						self.OP("JR", [self.DM_E], term = True)
					else:
						self.OP("JR", [self.DM_E], cc = self.cc_tab[y - 4], term = True)
				elif z == 1:
					if q:
						self.OP("ADD", ["HL", self.r16_tab[p]])
						self.reg.set16(res.p[0], None)
					else:
						self.OP("LD", [self.r16_tab[p], self.DM_W])
						self.reg.set16(res.p[0], res.dval)
				elif z == 2:
					if q:
						self.OP("LD", (["A", "(BC)"], ["A", "(DE)"], ["HL", self.DM_L], ["A", self.DM_L])[p])
						if p == 2: self.reg.set16(res.p[0], None)
						else: self.reg.set8(res.p[0], None)
					else:
						self.OP("LD", (["(BC)", "A"], ["(DE)", "A"], [self.DM_S, "HL"], [self.DM_S, "A"])[p])
				elif z == 3:
					self.OP("DEC" if q else "INC", [self.r16_tab[p]])
				elif z == 4:
					self.OP("INC", [self.r8_tab[y]])
					self.reg.add8(res.p[0], 1)
				elif z == 5:
					self.OP("DEC", [self.r8_tab[y]])
					self.reg.add8(res.p[0], -1)
				elif z == 6: # LD r8,n    (8-bit reg load with imm data)
					self.OP("LD", [self.r8_tab[y], self.DM_B])
					self.reg.set8(res.p[0], res.dval)
				elif z == 7:
					res.op = ("RLCA", "RRCA", "RLA", "RRA", "DAA", "CPL", "SCF", "CCF")[y]
					self.reg.set8("A", None)
			elif x == 1:  # the "LD R1,R2" group of opcodes
				if o == 0x76:
					res.op = "HALT" # halt is an exception: it's in the "LD R1,R2" group, but it's totally different for real
				else:
					self.OP("LD", [self.r8_tab[y], self.r8_tab[z]])
					self.reg.copy8(target = res.p[0], source = res.p[1])
			elif x == 2: # the "ALU" group of opcodes
				self.OP(self.alu_tab[y], [self.r8_tab[z]])
				if res.op in ("XOR", "SUB") and res.p[0] == "A":
					self.reg.set8("A", 0)
				else:
					self.reg.set8("A", None)
			elif x == 3:
				if z == 0:
					self.OP("RET", cc = self.cc_tab[y], term = True)  # conditional RET
				elif z == 1:
					if q:
						if p == 0: self.OP("RET", term = True)
						elif p == 1: self.OP("EXX")
						elif p == 2: self.OP("JP", ["HL"], term = True) # We don't know the address ... NOTE: JP (HL) format is invalid for real and a mistake of assemblers, it should be JP HL, as no indirect addressing!
						elif p == 3: self.OP("LD", ["SP", "HL"])
					else:
						self.OP("POP", [self.r16_tab2[p]])
						self.reg.set16(res.p[0], None)
				elif z == 2:
					self.OP("JP", [self.DM_J], cc = self.cc_tab[y], term = True) # conditional JP
				elif z == 3:
					if y == 0: self.OP("JP", [self.DM_J], term = True)
					elif y == 2: # y=1 case is 0xCB, handled as prefix already, so we are y==2 now
						self.OP("OUT", [self.DM_I, "A"])
					elif y == 3:
						self.OP("IN", ["A", self.DM_I])
						self.reg.set8("A", None)
					elif y == 4:
						self.OP("EX", ["(SP)", "HL"])
						self.reg.set16(res.p[1], None)
					elif y == 5:
						self.OP("EX", ["DE", "HL"])
						self.reg.ex16(res.p[0], res.p[1])
					elif y == 6:
						res.op = "DI"
					elif y == 7:
						res.op = "EI"
				elif z == 4:
					self.OP("CALL", [self.DM_J], cc = self.cc_tab[y])
				elif z == 5:
					if q: self.OP("CALL", [self.DM_J]) # for real, there are more bytes for this case, but DD/ED/FD is handled as "prefixes" at the beginning
					else: self.OP("PUSH", [self.r16_tab2[p]])
				elif z == 6:
					self.OP(self.alu_tab[y], [self.DM_B])
					self.reg.set8("A", None)
				elif z == 7:
					res.op, res.p, res.dval = "RST", [self.DM_R], y * 8

		else:
			raise RuntimeError("Internal problem: optab value is not handled: " + self.errbase)

		if res.op is None: # if res.op is not filled, it's a problem
			raise RuntimeError("Unknown " + self.errbase)
		return res # return with the result



if __name__ == "__main__": raise RuntimeError("Do not run this file directly, it's not the main 'executable'!")
